// Graphics library by ladyada/adafruit with init code from Rossum 
// MIT license

#include "ST7735.h"

#include "pins_arduino.h"
#include "wiring_private.h"
#include <SPI.h>

ST7735::ST7735(uint8_t cs, uint8_t rs, uint8_t sid, 
  uint8_t sclk, uint8_t rst) {
  _cs = cs;
  _rs = rs;
  _sid = sid;
  _sclk = sclk;
  _rst = rst;
  _backgroundColor=WHITE;
}

ST7735::ST7735(uint8_t cs, uint8_t rs,  uint8_t rst) {
  _cs = cs;
  _rs = rs;
  _sid = 0;
  _sclk = 0;
  _rst = rst;
  _backgroundColor=WHITE;
}


inline void ST7735::spiwrite(uint8_t c) {

  //Serial.println(c, HEX);

  if (!_sid) {
    SPI.transfer(c);
    return;
  }

  volatile uint8_t *sclkportreg = portOutputRegister(sclkport);
  volatile uint8_t *sidportreg = portOutputRegister(sidport);

  int8_t i;

  *sclkportreg |= sclkpin;

  for (i=7; i>=0; i--) {
    *sclkportreg &= ~sclkpin;
    //SCLK_PORT &= ~_BV(SCLK);
    
    if (c & _BV(i)) {
      *sidportreg |= sidpin;
      //digitalWrite(_sid, HIGH);
      //SID_PORT |= _BV(SID);
    } else {
      *sidportreg &= ~sidpin;
      //digitalWrite(_sid, LOW);
      //SID_PORT &= ~_BV(SID);
    }
    
    *sclkportreg |= sclkpin;
    //SCLK_PORT |= _BV(SCLK);
  }
}


void ST7735::writecommand(uint8_t c) {
  *portOutputRegister(rsport) &= ~ rspin;
  //digitalWrite(_rs, LOW);

  *portOutputRegister(csport) &= ~ cspin;
  //digitalWrite(_cs, LOW);

  //Serial.print("C ");
  spiwrite(c);

  *portOutputRegister(csport) |= cspin;
  //digitalWrite(_cs, HIGH);
}


void ST7735::writedata(uint8_t c) {
  *portOutputRegister(rsport) |= rspin;
  //digitalWrite(_rs, HIGH);

  *portOutputRegister(csport) &= ~ cspin;
    //digitalWrite(_cs, LOW);

  //Serial.print("D ");
  spiwrite(c);

  *portOutputRegister(csport) |= cspin;
  //digitalWrite(_cs, HIGH);
} 


void ST7735::setAddrWindow(uint8_t x0, uint8_t y0, uint8_t x1, uint8_t y1)
{
  writecommand(ST7735_CASET);  // column addr set
  writedata(0x00);
  writedata(x0+0);   // XSTART 
  writedata(0x00);
  writedata(x1+0);   // XEND

  writecommand(ST7735_RASET);  // row addr set
  writedata(0x00);
  writedata(y0+0);    // YSTART
  writedata(0x00);
  writedata(y1+0);    // YEND

  writecommand(ST7735_RAMWR);  // write to RAM
}

void ST7735::pushColor(uint16_t color) {
  *portOutputRegister(rsport) |= rspin;
  *portOutputRegister(csport) &= ~ cspin;

  spiwrite(color >> 8);    
  spiwrite(color);   

  *portOutputRegister(csport) |= cspin;
}

void ST7735::setBackgroundColor(uint16_t color)
{
  _backgroundColor=color;
}

void ST7735::drawPixel(uint8_t x, uint8_t y,uint16_t color) {

  if ((x >= width) || (y >= height)) return;

  setAddrWindow(x,y,x+1,y+1);

  // setup for data
  *portOutputRegister(rsport) |= rspin;
  *portOutputRegister(csport) &= ~ cspin;

  spiwrite(color >> 8);    
  spiwrite(color);   

  *portOutputRegister(csport) |= cspin;

}


void ST7735::fillScreen(uint16_t color) {

  setAddrWindow(0, 0, width-1, height-1);

  // setup for data
  *portOutputRegister(rsport) |= rspin;
  *portOutputRegister(csport) &= ~ cspin;

  for (uint8_t x=0; x < width; x++) {
    for (uint8_t y=0; y < height; y++) {
      spiwrite(color >> 8);    
      spiwrite(color);    
    }
  }

  *portOutputRegister(csport) |= cspin;
}

void ST7735::initB(void) {
  // set pin directions
  pinMode(_rs, OUTPUT);

  if (_sclk) {
    pinMode(_sclk, OUTPUT);
    sclkport = digitalPinToPort(_sclk);
    sclkpin = digitalPinToBitMask(_sclk);

    pinMode(_sid, OUTPUT);
    sidport = digitalPinToPort(_sid);
    sidpin = digitalPinToBitMask(_sid);
  } else {
    // using the hardware SPI
    // has to be started external
    /*SPI.begin();
    SPI.setClockDivider(SPI_CLOCK_DIV2);
    SPI.setDataMode(SPI_MODE3);*/
  }
  // toggle RST low to reset; CS low so it'll listen to us
  pinMode(_cs, OUTPUT);
  digitalWrite(_cs, LOW);
  cspin = digitalPinToBitMask(_cs);
  csport = digitalPinToPort(_cs);

  rspin = digitalPinToBitMask(_rs);
  rsport = digitalPinToPort(_rs);

  if (_rst) {
    pinMode(_rst, OUTPUT);
    digitalWrite(_rst, HIGH);
    delay(500);
    digitalWrite(_rst, LOW);
    delay(500);
    digitalWrite(_rst, HIGH);
    delay(500);
  }

  writecommand(ST7735_SWRESET); // software reset
  delay(50);
  writecommand(ST7735_SLPOUT);  // out of sleep mode
  delay(500);
  
  writecommand(ST7735_COLMOD);  // set color mode
  writedata(0x05);        // 16-bit color
  delay(10);
  
  writecommand(ST7735_FRMCTR1);  // frame rate control
  writedata(0x00);  // fastest refresh
  writedata(0x06);  // 6 lines front porch
  writedata(0x03);  // 3 lines backporch
  delay(10);
  
  writecommand(ST7735_MADCTL);  // memory access control (directions)
  writedata(0x00);  // row address/col address, bottom to top refresh
  madctl = 0x00;

  writecommand(ST7735_DISSET5);  // display settings #5
  writedata(0x15);  // 1 clock cycle nonoverlap, 2 cycle gate rise, 3 cycle oscil. equalize
  writedata(0x02);  // fix on VTL
 
  writecommand(ST7735_INVCTR);  // display inversion control
  writedata(0x0);  // line inversion
 
  writecommand(ST7735_PWCTR1);  // power control
  writedata(0x02);      // GVDD = 4.7V 
  writedata(0x70);      // 1.0uA
  delay(10);
  writecommand(ST7735_PWCTR2);  // power control
  writedata(0x05);      // VGH = 14.7V, VGL = -7.35V 
  writecommand(ST7735_PWCTR3);  // power control
  writedata(0x01);      // Opamp current small 
  writedata(0x02);      // Boost frequency
  
  
  writecommand(ST7735_VMCTR1);  // power control
  writedata(0x3C);      // VCOMH = 4V
  writedata(0x38);      // VCOML = -1.1V
  delay(10);
  
  writecommand(ST7735_PWCTR6);  // power control
  writedata(0x11); 
  writedata(0x15);
  
  writecommand(ST7735_GMCTRP1);
  writedata(0x0f);	//writedata(0x09);  
  writedata(0x1a);  //writedata(0x16);
  writedata(0x0f);  //writedata(0x09);
  writedata(0x18);  //writedata(0x20);
  writedata(0x2f);  //writedata(0x21);
  writedata(0x28);  //writedata(0x1B);
  writedata(0x20);  //writedata(0x13);
  writedata(0x22);  //writedata(0x19);
  writedata(0x1f);  //writedata(0x17);
  writedata(0x1b);  //writedata(0x15);
  writedata(0x23);  //writedata(0x1E);
  writedata(0x37);  //writedata(0x2B);
  writedata(0x00);  //writedata(0x04);
  writedata(0x07);  //writedata(0x05);
  writedata(0x02);  //writedata(0x02);
  writedata(0x10);  //writedata(0x0E);
  writecommand(ST7735_GMCTRN1);
  writedata(0x0f);   //writedata(0x0B); 
  writedata(0x1b);   //writedata(0x14); 
  writedata(0x0f);   //writedata(0x08); 
  writedata(0x17);   //writedata(0x1E); 
  writedata(0x33);   //writedata(0x22); 
  writedata(0x2c);   //writedata(0x1D); 
  writedata(0x29);   //writedata(0x18); 
  writedata(0x2e);   //writedata(0x1E); 
  writedata(0x30);   //writedata(0x1B); 
  writedata(0x30);   //writedata(0x1A); 
  writedata(0x39);   //writedata(0x24); 
  writedata(0x3f);   //writedata(0x2B); 
  writedata(0x00);   //writedata(0x06); 
  writedata(0x07);   //writedata(0x06); 
  writedata(0x03);   //writedata(0x02); 
  writedata(0x10);   //writedata(0x0F); 
  delay(10);
  
  writecommand(ST7735_CASET);  // column addr set
  writedata(0x00);
  writedata(0x02);   // XSTART = 2
  writedata(0x00);
  writedata(0x81);   // XEND = 129

  writecommand(ST7735_RASET);  // row addr set
  writedata(0x00);
  writedata(0x02);    // XSTART = 1
  writedata(0x00);
  writedata(0x81);    // XEND = 160

  writecommand(ST7735_NORON);  // normal display on
  delay(10);
  
  writecommand(ST7735_RAMWR);
  delay(500);
  
  writecommand(ST7735_DISPON);
  delay(500);
}



void ST7735::initR(void) {
  // set pin directions
  pinMode(_rs, OUTPUT);

  if (_sclk) {
    pinMode(_sclk, OUTPUT);
    sclkport = digitalPinToPort(_sclk);
    sclkpin = digitalPinToBitMask(_sclk);

    pinMode(_sid, OUTPUT);
    sidport = digitalPinToPort(_sid);
    sidpin = digitalPinToBitMask(_sid);
  } else {
    // using the hardware SPI
    // has to be enabled external
    /*SPI.begin();
    SPI.setClockDivider(SPI_CLOCK_DIV2);
    SPI.setDataMode(SPI_MODE3);*/
  }
  // toggle RST low to reset; CS low so it'll listen to us
  pinMode(_cs, OUTPUT);
  digitalWrite(_cs, LOW);
  cspin = digitalPinToBitMask(_cs);
  csport = digitalPinToPort(_cs);

  rspin = digitalPinToBitMask(_rs);
  rsport = digitalPinToPort(_rs);

  if (_rst) {
    pinMode(_rst, OUTPUT);
    digitalWrite(_rst, HIGH);
    delay(500);
    digitalWrite(_rst, LOW);
    delay(500);
    digitalWrite(_rst, HIGH);
    delay(500);
  }

 writecommand(ST7735_SWRESET); // software reset
  delay(150);

  writecommand(ST7735_SLPOUT);  // out of sleep mode
  delay(500);

  writecommand(ST7735_FRMCTR1);  // frame rate control - normal mode
  writedata(0x01);  // frame rate = fosc / (1 x 2 + 40) * (LINE + 2C + 2D)
  writedata(0x2C); 
  writedata(0x2D); 

  writecommand(ST7735_FRMCTR2);  // frame rate control - idle mode
  writedata(0x01);  // frame rate = fosc / (1 x 2 + 40) * (LINE + 2C + 2D)
  writedata(0x2C); 
  writedata(0x2D); 

  writecommand(ST7735_FRMCTR3);  // frame rate control - partial mode
  writedata(0x01); // dot inversion mode
  writedata(0x2C); 
  writedata(0x2D); 
  writedata(0x01); // line inversion mode
  writedata(0x2C); 
  writedata(0x2D); 
  
  writecommand(ST7735_INVCTR);  // display inversion control
  writedata(0x07);  // no inversion

  writecommand(ST7735_PWCTR1);  // power control
  writedata(0xA2);      
  writedata(0x02);      // -4.6V
  writedata(0x84);      // AUTO mode

  writecommand(ST7735_PWCTR2);  // power control
  writedata(0xC5);      // VGH25 = 2.4C VGSEL = -10 VGH = 3 * AVDD

  writecommand(ST7735_PWCTR3);  // power control
  writedata(0x0A);      // Opamp current small 
  writedata(0x00);      // Boost frequency

  writecommand(ST7735_PWCTR4);  // power control
  writedata(0x8A);      // BCLK/2, Opamp current small & Medium low
  writedata(0x2A);     

  writecommand(ST7735_PWCTR5);  // power control
  writedata(0x8A);    
  writedata(0xEE);     

  writecommand(ST7735_VMCTR1);  // power control
  writedata(0x0E);  

  writecommand(ST7735_INVOFF);    // don't invert display

  writecommand(ST7735_MADCTL);  // memory access control (directions)
  writedata(0xC0);  // row address/col address, bottom to top refresh, RGB
  madctl = 0xC0;
  
  writecommand(ST7735_COLMOD);  // set color mode
  writedata(0x05);        // 16-bit color

  writecommand(ST7735_CASET);  // column addr set
  writedata(0x00);
  writedata(0x00);   // XSTART = 0
  writedata(0x00);
  writedata(0x7F);   // XEND = 127

  writecommand(ST7735_RASET);  // row addr set
  writedata(0x00);
  writedata(0x00);    // XSTART = 0
  writedata(0x00);
  writedata(0x9F);    // XEND = 159

  writecommand(ST7735_GMCTRP1);
  writedata(0x0f);
  writedata(0x1a);
  writedata(0x0f);
  writedata(0x18);
  writedata(0x2f);
  writedata(0x28);
  writedata(0x20);
  writedata(0x22);
  writedata(0x1f);
  writedata(0x1b);
  writedata(0x23);
  writedata(0x37);
  writedata(0x00);
  writedata(0x07);
  writedata(0x02);
  writedata(0x10);
  writecommand(ST7735_GMCTRN1);
  writedata(0x0f); 
  writedata(0x1b); 
  writedata(0x0f); 
  writedata(0x17); 
  writedata(0x33); 
  writedata(0x2c); 
  writedata(0x29); 
  writedata(0x2e); 
  writedata(0x30); 
  writedata(0x30); 
  writedata(0x39); 
  writedata(0x3f); 
  writedata(0x00); 
  writedata(0x07); 
  writedata(0x03); 
  writedata(0x10); 
  
  writecommand(ST7735_DISPON);
  delay(100);

  writecommand(ST7735_NORON);  // normal display on
  delay(10);
}

// fill a circle
void ST7735::fillCircle(uint8_t x0, uint8_t y0, uint8_t r, uint16_t color) {
  int16_t f = 1 - r;
  int16_t ddF_x = 1;
  int16_t ddF_y = -2 * r;
  int16_t x = 0;
  int16_t y = r;

  drawVerticalLine(x0, y0-r, 2*r+1, color);

  while (x<y) {
    if (f >= 0) {
      y--;
      ddF_y += 2;
      f += ddF_y;
    }
    x++;
    ddF_x += 2;
    f += ddF_x;
  
    drawVerticalLine(x0+x, y0-y, 2*y+1, color);
    drawVerticalLine(x0-x, y0-y, 2*y+1, color);
    drawVerticalLine(x0+y, y0-x, 2*x+1, color);
    drawVerticalLine(x0-y, y0-x, 2*x+1, color);
  }
}

// draw a circle outline
void ST7735::drawCircle(uint8_t x0, uint8_t y0, uint8_t r, 
			uint16_t color) {
  int16_t f = 1 - r;
  int16_t ddF_x = 1;
  int16_t ddF_y = -2 * r;
  int16_t x = 0;
  int16_t y = r;

  drawPixel(x0, y0+r, color);
  drawPixel(x0, y0-r, color);
  drawPixel(x0+r, y0, color);
  drawPixel(x0-r, y0, color);

  while (x<y) {
    if (f >= 0) {
      y--;
      ddF_y += 2;
      f += ddF_y;
    }
    x++;
    ddF_x += 2;
    f += ddF_x;
  
    drawPixel(x0 + x, y0 + y, color);
    drawPixel(x0 - x, y0 + y, color);
    drawPixel(x0 + x, y0 - y, color);
    drawPixel(x0 - x, y0 - y, color);
    
    drawPixel(x0 + y, y0 + x, color);
    drawPixel(x0 - y, y0 + x, color);
    drawPixel(x0 + y, y0 - x, color);
    drawPixel(x0 - y, y0 - x, color);
    
  }
}

uint8_t ST7735::getRotation() {
  return madctl;
}

void ST7735::setRotation(uint8_t m) {
  madctl = m;
  writecommand(ST7735_MADCTL);  // memory access control (directions)
  writedata(madctl);  // row address/col address, bottom to top refresh
}

// draw a rectangle
void ST7735::drawRect(uint8_t x, uint8_t y, uint8_t w, uint8_t h, 
		      uint16_t color) {
  // smarter version
  drawHorizontalLine(x, y, w, color);
  drawHorizontalLine(x, y+h-1, w, color);
  drawVerticalLine(x, y, h, color);
  drawVerticalLine(x+w-1, y, h, color);
}

void ST7735::fillRect(uint8_t x, uint8_t y, uint8_t w, uint8_t h, 
		      uint16_t color) {
  // smarter version

  setAddrWindow(x, y, x+w-1, y+h-1);

  // setup for data
  digitalWrite(_rs, HIGH);
  digitalWrite(_cs, LOW);

  // display controller writes first in X and increments Y after window width
  for (y=0; y < h; y++) {
    for (x=0; x < w; x++) {
      spiwrite(color >> 8);    
      spiwrite(color);    
    }
  }
  digitalWrite(_cs, HIGH);
}

void ST7735::drawVerticalLine(uint8_t x, uint8_t y, uint8_t length, uint16_t color)
{
  if (x >= width) return;
  if (y+length >= height) length = height-y-1;

  drawFastLine(x,y,length,color,1);
}

void ST7735::drawHorizontalLine(uint8_t x, uint8_t y, uint8_t length, uint16_t color)
{
  if (y >= height) return;
  if (x+length >= width) length = width-x-1;

  drawFastLine(x,y,length,color,0);
}

void ST7735::drawFastLine(uint8_t x, uint8_t y, uint8_t length, 
			  uint16_t color, uint8_t rotflag)
{
  if (rotflag) {
    setAddrWindow(x, y, x, y+length);
  } else {
    setAddrWindow(x, y, x+length, y+1);
  }
  // setup for data
  digitalWrite(_rs, HIGH);
  digitalWrite(_cs, LOW);

  while (length--) {
    spiwrite(color >> 8);    
    spiwrite(color);    
  }
  digitalWrite(_cs, HIGH);
}


// bresenham's algorithm - thx wikpedia
void ST7735::drawLine(int16_t x0, int16_t y0, int16_t x1, int16_t y1, 
		      uint16_t color) {
  uint16_t steep = abs(y1 - y0) > abs(x1 - x0);
  if (steep) {
    swap(x0, y0);
    swap(x1, y1);
  }

  if (x0 > x1) {
    swap(x0, x1);
    swap(y0, y1);
  }

  uint16_t dx, dy;
  dx = x1 - x0;
  dy = abs(y1 - y0);

  int16_t err = dx / 2;
  int16_t ystep;

  if (y0 < y1) {
    ystep = 1;
  } else {
    ystep = -1;}

  for (; x0<=x1; x0++) {
    if (steep) {
      drawPixel(y0, x0, color);
    } else {
      drawPixel(x0, y0, color);
    }
    err -= dy;
    if (err < 0) {
      y0 += ystep;
      err += dx;
    }
  }
}


/**
 * Draw a rectangle with rounder corners
 *
 * @param x the x coordinate of the upper left corner of the rectangle
 * @param y the y coordinate of the upper left corner of the rectangle
 * @param width width of the rectangle
 * @param height height of the rectangle
 * @param radius ????
 * @param 16-bit color default BLACK 
 *
 * Draws a rectangle the same as DrawRect() but with rounded corners.
 * Radius is a value from 1 to half the height or width of the rectangle.
 * (what does that really mean?????)
 * Which is it, or is it a the smaller of the two?
 * FIXME FIXME need more description here.
 *
 * @see DrawRect()
 */

void ST7735::drawRoundRect(uint8_t x, uint8_t y, uint8_t width, uint8_t height, uint8_t radius, uint16_t color=BLACK) 
{
  	int16_t tSwitch; 
	uint8_t x1 = 0, y1 = radius;
  	tSwitch = 3 - 2 * radius;
	
	while (x1 <= y1) {
	    this->drawPixel(x+radius - x1, y+radius - y1, color);
	    this->drawPixel(x+radius - y1, y+radius - x1, color);

	    this->drawPixel(x+width-radius + x1, y+radius - y1, color);
	    this->drawPixel(x+width-radius + y1, y+radius - x1, color);
	    
	    this->drawPixel(x+width-radius + x1, y+height-radius + y1, color);
	    this->drawPixel(x+width-radius + y1, y+height-radius + x1, color);

	    this->drawPixel(x+radius - x1, y+height-radius + y1, color);
	    this->drawPixel(x+radius - y1, y+height-radius + x1, color);

	    if (tSwitch < 0) {
	    	tSwitch += (4 * x1 + 6);
	    } else {
	    	tSwitch += (4 * (x1 - y1) + 10);
	    	y1--;
	    }
	    x1++;
	}
	  	
	this->drawHorizontalLine(x+radius, y, width-(2*radius), color);		// top
	this->drawHorizontalLine(x+radius, y+height, width-(2*radius), color);	// bottom
	this->drawVerticalLine(x, y+radius, height-(2*radius), color);		// left
	this->drawVerticalLine(x+width, y+radius, height-(2*radius), color);	// right
}



void ST7735::selectFont(Font_t font)
{
	this->Font = font;
}

uint8_t ST7735::getFontHeight()
{
  return( FontRead(this->Font+FONT_HEIGHT) );
}


uint8_t ST7735::putChar(uint8_t x, const uint8_t y, uint8_t c, const uint16_t color=BLACK)
// returns x position after printed character
// function is a little bit speed optimized, with AT328p 16MHz: 
// 0.61ms for 5x7 character
// 1.26ms for a 7x15 character 
// 3.55ms for a 10x24 character
{
  if (this->Font == 0)
    return 0; // no font selected

  /*
  * check for special character processing
  */

  if (c < 0x20)  // special char, do nothing
  {
    return x;
  }
	   
  uint8_t fontWidth = 0;
  uint8_t FontHeight = FontRead(this->Font+FONT_HEIGHT);
  uint8_t bytes = (FontHeight+7)/8; /* calculates FontHeight in rounded up bytes */

  uint8_t firstChar = FontRead(this->Font+FONT_FIRST_CHAR);
  uint8_t charCount = FontRead(this->Font+FONT_CHAR_COUNT);
	
  uint16_t index = 0;
  uint8_t thielefont;

  if (c < firstChar || c >= (firstChar+charCount)) {
    return x; // invalid char
  }
  c-= firstChar;

  if ( isFixedWidthFont(this->Font) {
    thielefont = 0;
    fontWidth = FontRead(this->Font+FONT_FIXED_WIDTH); 
    index = c*bytes*fontWidth+FONT_WIDTH_TABLE;
  }
  else  {
    // variable width font, read width data, to get the index
    thielefont = 1;
    /*
     * Because there is no table for the offset of where the data
     * for each character glyph starts, run the table and add up all the
     * widths of all the characters prior to the character we
     * need to locate.
     */
    for (uint8_t i=0; i<c; i++) {  
      index += FontRead(this->Font+FONT_WIDTH_TABLE+i);
    }
    /*
     * Calculate the offset of where the font data
     * for our character starts.
     * The index value from above has to be adjusted because
     * there is potentialy more than 1 byte per column in the glyph,
     * when the characgter is taller than 8 bits.
     * To account for this, index has to be multiplied
     * by the FontHeight in bytes because there is one byte of font
     * data for each vertical 8 pixels.
     * The index is then adjusted to skip over the font fontWidth data
     * and the font header information.
     */
     index = index*bytes+charCount+FONT_WIDTH_TABLE;
     /*
      * Finally, fetch the fontWidth of our character
      */
     fontWidth = FontRead(this->Font+FONT_WIDTH_TABLE+c);
  }  // end else

  // if character does not fit to the screen, write it at the end (no line wrap)
  if (x + fontWidth >= this->width)
  {
    x=width-fontWidth;
  }

  // last but not least, draw the character
  setAddrWindow(x, y, x+fontWidth-1, y+FontHeight-1);

  // setup display for data transmission
  digitalWrite(_rs, HIGH);
  digitalWrite(_cs, LOW);

  /*
   * Need to write to Display Memory in X direction
   * Y automatically increments due to set window size
   * Because GLCD fonts are organized in Y direction, need some ugly 
   * Font Data accesses to map the Font to the display memory
   * GLCS has stored the font in the following way:
   * Each byte is one X-step and eight Y-Steps. First X/Y pixel is given by the LSB
   * After all Bytes of one X line (font width) the next Y+8 pixel are coming
   * if less than 8-bit are used in a byte the LSB's are not used
   */

  for (uint8_t i=0; i<bytes; i++)	/* each vertical byte */
  { 
    uint16_t page = i*fontWidth; // page must be 16 bit to prevent overflow		
    uint8_t bitpnt=0x01;    // pointer to the bit to be drawn
    uint8_t nr_of_pixel=8;  // initial value for a byte
    const uint8_t* tmpFontPnt=this->Font+index+page;

    /* is it last byte of multibyte tall font? or a very small font? */
    uint8_t lastByte=( FontHeight > 8 && FontHeight < (i+1)*8 ) || (FontHeight < 8);
    if ( lastByte )
    /* is it last byte of multibyte tall font? or a very small font? */
    {
      nr_of_pixel=FontHeight-i*8;
      if (FontHeight<8)  // Workaround: bug in font-data for small fonts...
         bitpnt <<= (i+1)*8-FontHeight-1;        // remove unused bits
       else
         bitpnt <<= (i+1)*8-FontHeight;        
    }
    for (uint8_t p=0; p<nr_of_pixel; p++)  {
      uint8_t yp = i*8+p;
      for (uint8_t j=0; j<fontWidth; j++)  {   // write in x-direction
         uint8_t data = FontRead(tmpFontPnt+j);
         if (data & bitpnt)  {
           spiwrite(color >> 8);    
           spiwrite(color);    
           //drawPixel(x+j, y+yp, color);
         }
         else  {
           spiwrite(_backgroundColor >> 8);    
           spiwrite(_backgroundColor);    
           //drawPixel(x+j, y+yp, _backgroundColor);
         }
       }
       bitpnt <<= 1;
     }  // end  for p
  }  // end for i
  digitalWrite(_cs, HIGH);
  return (x+fontWidth); // valid char, return new x position
}



uint8_t ST7735::putCharOld(uint8_t x, uint8_t y, uint8_t c, uint16_t color=BLACK)
// returns x position after printed character
// Not speed optimized: 15.8ms for a 8x14 character with AT328p 16MHz
{
    if (this->Font == 0)
	  return 0; // no font selected

	/*
	 * check for special character processing
	 */

	if (c < 0x20)  // special char, do nothing
	{
		return x;
	}
	   
	uint8_t fontWidth = 0;
	uint8_t FontHeight = FontRead(this->Font+FONT_HEIGHT);
	uint8_t bytes = (FontHeight+7)/8; /* calculates FontHeight in rounded up bytes */
	
	uint8_t firstChar = FontRead(this->Font+FONT_FIRST_CHAR);
	uint8_t charCount = FontRead(this->Font+FONT_CHAR_COUNT);
	
	uint16_t index = 0;
	uint8_t thielefont;

	if (c < firstChar || c >= (firstChar+charCount)) {
		return x; // invalid char
	}
	c-= firstChar;

	if ( isFixedWidthFont(this->Font) {
		thielefont = 0;
		fontWidth = FontRead(this->Font+FONT_FIXED_WIDTH); 
		index = c*bytes*fontWidth+FONT_WIDTH_TABLE;
	}
	else  {
	// variable width font, read width data, to get the index
		thielefont = 1;
		/*
		 * Because there is no table for the offset of where the data
		 * for each character glyph starts, run the table and add up all the
		 * widths of all the characters prior to the character we
		 * need to locate.
		 */
	   for (uint8_t i=0; i<c; i++) {  
		 index += FontRead(this->Font+FONT_WIDTH_TABLE+i);
	   }
		/*
		 * Calculate the offset of where the font data
		 * for our character starts.
		 * The index value from above has to be adjusted because
		 * there is potentialy more than 1 byte per column in the glyph,
		 * when the characgter is taller than 8 bits.
		 * To account for this, index has to be multiplied
		 * by the FontHeight in bytes because there is one byte of font
		 * data for each vertical 8 pixels.
		 * The index is then adjusted to skip over the font fontWidth data
		 * and the font header information.
		 */

	   index = index*bytes+charCount+FONT_WIDTH_TABLE;

		/*
		 * Finally, fetch the fontWidth of our character
		 */
	   fontWidth = FontRead(this->Font+FONT_WIDTH_TABLE+c);
    }

	/*
	 * If the character won't fit in the text area,
	 * overwrite the end of the line with new char
	 */
	if (x + fontWidth >= this->width)
	{
	  x=width-fontWidth;
	}

	// last but not least, draw the character

	/*
	 * Draw each column of the glyph (character) horizontally
	 * 8 bits (1 page) at a time.
	 * i.e. if a font is taller than 8 bits, draw upper 8 bits first,
	 * Then drop down and draw next 8 bits and so on, until done.
	 * This code depends on WriteData() doing writes that span LCD
	 * memory pages, which has issues because the font data isn't
	 * always a multiple of 8 bits.
	 */

	for (uint8_t i=0; i<bytes; i++)	/* each vertical byte */
	{
		uint16_t page = i*fontWidth; // page must be 16 bit to prevent overflow
		for (uint8_t j=0; j<fontWidth; j++) /* each column */
		{
			uint8_t data = FontRead(this->Font+index+page+j);
		
			/*
			 * This funkyness is because when the character glyph is not a
			 * multiple of 8 in FontHeight, the residual bits in the font data
			 * were aligned to the incorrect end of the byte with respect 
			 * to the GLCD. I believe that this was an initial oversight (bug)
			 * in Thieles font creator program. It is easily fixed
			 * in the font program but then creates a potential backward
			 * compatiblity problem.
			 *	--- bperrybap
			 */

			uint8_t nr_of_pixel=8;
			if ( (FontHeight > 8 && FontHeight < (i+1)*8) || (FontHeight < 8)  )
			/* is it last byte of multibyte tall font? or a very small font? */
			{
				data >>= (i+1)*8-FontHeight;        // scroll remaining bits to the right
				nr_of_pixel=FontHeight-i*8;
			}
			for (uint8_t p=i*8; p<i*8+nr_of_pixel; p++)  {
			  if (data & 0x1)  {
 			    drawPixel(x+j, y+p, color);
                          }
                          else  {
 			    drawPixel(x+j, y+p, _backgroundColor);
                          }
			  data >>= 1;
			}

		}  // end  for j

	}  // end for i

	return (x+fontWidth); // valid char, return new x position
}


uint8_t ST7735::putS(uint8_t x, uint8_t y, char *str, uint16_t color=BLACK)
{
  uint8_t first=1;
  uint8_t fontHeight = FontRead(this->Font+FONT_HEIGHT);

  while(*str)  {
    if (!first)  {
      // space between characters
      drawVerticalLine(x,y,fontHeight, _backgroundColor);
      x++;
    }
    x=this->putChar(x, y, (uint8_t)*str, color);
    str++;
    first=0;
  }
  return(x);
}


/**
 * output a String class string
 *
 * @param str String class string
 *
 * Outputs all the characters in the string to the text area. 
 * See PutChar() for a full description of how characters are
 * written to the text area.
 *
 * @see PutChar()
 * @see Puts_P()
 * @see DrawString()
 * @see DrawString_P()
 * @see write()
 */
uint8_t  ST7735::putS(uint8_t x, uint8_t y, const String &str, uint16_t color=BLACK)
{
  uint8_t fontHeight = FontRead(this->Font+FONT_HEIGHT);

  for (uint8_t i = 0; i < str.length(); i++)  {
    if (i!=0)  {
      // space between characters
      drawVerticalLine(x,y,fontHeight, _backgroundColor);
      x++;
    }
    x=this->putChar(x, y, str[i], color);
  }
  return(x);
}

/**
 * output a program memory character string
 *
 * @param str pointer to a null terminated character string stored in program memory
 *
 * Outputs all the characters in the string to the text area. 
 * See PutChar() for a full description of how characters are
 * written to the text area.
 *
 * @see PutChar()
 * @see Puts()
 * @see DrawString()
 * @see DrawString_P()
 * @see write()
 */

uint8_t ST7735::putS_P(uint8_t x, uint8_t y, PGM_P str, uint16_t color=BLACK)
{
  uint8_t c;
  uint8_t first=1;
  uint8_t fontHeight = FontRead(this->Font+FONT_HEIGHT);

    while((c = pgm_read_byte(str)) != 0)  {
      if (!first)  {
        // space between characters
        drawVerticalLine(x,y,fontHeight, _backgroundColor);
        x++;
      }
      x=this->putChar(x, y, c, color);
      str++;
      first=0;
    }
  return(x);
}

/**
 * output a character string at x,y coordinate
 *
 * @param str String class string
 * @param x specifies the horizontal location
 * @param y specifies the vertical location
 *
 *
 * Outputs all the characters in the string to the text area. 
 * X & Y are zero based pixel coordinates and are relative to 
 * the upper left corner of the text area.
 *
 * See PutChar() for a full description of how characters are
 * written to the text area.
 *
 *
 * @see PutChar()
 * @see Puts()
 * @see Puts_P()
 * @see DrawString_P()
 * @see write()
 */



/**
 * Returns the pixel width of a character
 *
 * @param c character to be sized
 *
 * @return The width in pixels of the given character
 * including any inter-character gap pixels following the character when the character is
 * rendered on the display.
 *
 * @note The font for the character is the most recently selected font.
 *
 * @see StringWidth()
 * @see StringWidth_P()
 */

uint8_t ST7735::charWidth(uint8_t c)
{
	uint8_t width = 0;
	
    if(isFixedWidthFont(this->Font){
		width = FontRead(this->Font+FONT_FIXED_WIDTH)+1;  // there is 1 pixel pad here
	} 
    else{ 
	    // variable width font 
		uint8_t firstChar = FontRead(this->Font+FONT_FIRST_CHAR);
		uint8_t charCount = FontRead(this->Font+FONT_CHAR_COUNT);
	
		// read width data
		if(c >= firstChar && c < (firstChar+charCount)) {
			c -= firstChar;
			width = FontRead(this->Font+FONT_WIDTH_TABLE+c)+1;
		}
	}	
	return width;
}

/**
 * Returns the pixel width of a string
 *
 * @param str pointer to string stored in RAM
 *
 * @return the width in pixels of the sum of all the characters in the
 * the string pointed to by str. 
 *
 * @see CharWidth()
 * @see StringWidth_P()
 */

uint16_t ST7735::stringWidth(const char* str)
{
	uint16_t width = 0;
	
	while(*str != 0) {
		width += this->charWidth(*str++);
	}
	
	return width;
}

/**
 * Returns the pixel width of a character
 *
 * @param str pointer to string stored in program memory
 *
 * @return the width in pixels of the sum of all the characters in the
 * the string pointed to by str. 
 *
 * @see CharWidth()
 * @see StringWidth()
 */

uint16_t ST7735::stringWidth_P(PGM_P str)
{
	uint16_t width = 0;
	
	while(pgm_read_byte(str) != 0) {
		width += this->charWidth(pgm_read_byte(str++));
	}
	
	return width;
}

/**
 * Returns the pixel width of a character
 *
		 * @param str String class string
 *
 * @return the width in pixels of the sum of all the characters in the
 * the string pointed to by str. 
 *
 * @see CharWidth()
 * @see StringWidth()
 */

uint16_t ST7735::stringWidth(String &str)
{
	uint16_t width = 0;

	for (uint16_t i = 0; i < str.length(); i++)
	{
		width += this->charWidth(str[i]);
	}
	
	return width;
}



//////////
/*
uint8_t ST7735::spiread(void) {
  uint8_t r = 0;
  if (_sid > 0) {
    r = shiftIn(_sid, _sclk, MSBFIRST);
  } else {
    //SID_DDR &= ~_BV(SID);
    //int8_t i;
    //for (i=7; i>=0; i--) {
    //  SCLK_PORT &= ~_BV(SCLK);
    //  r <<= 1;
    //  r |= (SID_PIN >> SID) & 0x1;
    //  SCLK_PORT |= _BV(SCLK);
    //}
    //SID_DDR |= _BV(SID);
    
  }
  return r;
}



void ST7735::dummyclock(void) {

  if (_sid > 0) {
    digitalWrite(_sclk, LOW);
    digitalWrite(_sclk, HIGH);
  } else {
    // SCLK_PORT &= ~_BV(SCLK);
    //SCLK_PORT |= _BV(SCLK);
  }
}
uint8_t ST7735::readdata(void) {
  *portOutputRegister(rsport) |= rspin;

  *portOutputRegister(csport) &= ~ cspin;

  uint8_t r = spiread();

  *portOutputRegister(csport) |= cspin;

  return r;
} 

uint8_t ST7735::readcommand8(uint8_t c) {
  digitalWrite(_rs, LOW);

  *portOutputRegister(csport) &= ~ cspin;

  spiwrite(c);

  digitalWrite(_rs, HIGH);
  pinMode(_sid, INPUT); // input!
  digitalWrite(_sid, LOW); // low
  spiread();
  uint8_t r = spiread();


  *portOutputRegister(csport) |= cspin;


  pinMode(_sid, OUTPUT); // back to output
  return r;
}


uint16_t ST7735::readcommand16(uint8_t c) {
  digitalWrite(_rs, LOW);
  if (_cs)
    digitalWrite(_cs, LOW);

  spiwrite(c);
  pinMode(_sid, INPUT); // input!
  uint16_t r = spiread();
  r <<= 8;
  r |= spiread();
  if (_cs)
    digitalWrite(_cs, HIGH);

  pinMode(_sid, OUTPUT); // back to output
  return r;
}

uint32_t ST7735::readcommand32(uint8_t c) {
  digitalWrite(_rs, LOW);
  if (_cs)
    digitalWrite(_cs, LOW);
  spiwrite(c);
  pinMode(_sid, INPUT); // input!

  dummyclock();
  dummyclock();

  uint32_t r = spiread();
  r <<= 8;
  r |= spiread();
  r <<= 8;
  r |= spiread();
  r <<= 8;
  r |= spiread();
  if (_cs)
    digitalWrite(_cs, HIGH);

  pinMode(_sid, OUTPUT); // back to output
  return r;
}

*/
